module FP_Parser(parseProg, parseTerm) where

import Program
import Term
import ErrorMonad
import Text.ParserCombinators.Parsec
import Text.Parsec.Char
import Control.Monad(unless)
import qualified Data.Set as S

lexeme p = p <* spaces
lChar c = lexeme (char c)
lString s = lexeme (string s)
lString' s = lexeme (string' s)

arguments1 :: GenParser Char u a -> GenParser Char u [a]
arguments1 p = between (lChar '(') (lChar ')') (sepBy p (lChar ','))
  <|> ((\ a -> [a]) <$> p)
  <|> return []

arguments :: GenParser Char u a -> GenParser Char u [a]
arguments p = between (lChar '(') (lChar ')') (sepBy p (lChar ','))
  <|> return []

-- data definitions

constrName :: GenParser Char u FSym
constrName = lexeme (return (:) <*> upper <*> many alphaNum)

tyName :: GenParser Char u Type
tyName = constrName

constrDef :: GenParser Char u (FSym, [Type])
constrDef = return (,) <*> constrName <*> arguments1 tyName

dataDef :: GenParser Char u DataDefinition
dataDef = do
  _ <- lString "data"
  ty <- tyName
  _ <- lChar '='
  cs <- sepBy constrDef (lChar '|')
  return $ Data ty (map (\ (c,tyIn) -> (c,(tyIn,ty))) cs)


--- function definitions

funName :: GenParser Char u FSym
funName = lexeme (return (:) <*> lower <*> many alphaNum)

name :: GenParser Char u FSym
name = lexeme (return (:) <*> letter <*> many alphaNum)

varName :: GenParser Char u Var
varName = funName

inputTypes :: GenParser Char u [Type]
inputTypes = arguments1 tyName
  
functionType :: GenParser Char u FSymInfo
functionType = try (return (,) <*> inputTypes <* lexeme (string "->") <*> tyName)
  <|> (\ ty -> ([],ty)) <$> tyName
  
funTypeDef :: GenParser Char u (FSym, FSymInfo)
funTypeDef = return (,) <*> funName <* lString "::" <*> functionType

pattern :: GenParser Char u Term
pattern = (Var <$> varName)
  <|> (return Fun <*> constrName <*> arguments pattern)

vars (Var x) = S.insert x S.empty
vars (Fun f ts) = S.unions (map vars ts)


lhs :: FSym -> GenParser Char u Term
lhs f = return Fun <*> lString' f <*> arguments pattern

term :: S.Set Var -> GenParser Char u Term
term vs = (try (do {x <- varName; unless (S.member x vs) (fail "not a var"); return $ Var x}))
  <|> (return Fun <*> name <*> arguments (term vs))

equation :: FSym -> GenParser Char u (Term,Term)
equation f = do
  l <- lhs f
  _ <- lChar '='
  let vs = vars l
  r <- term vs
  return (l,r)
  
funDef :: GenParser Char u FunctionDefinition 
funDef = do
  (f, fty) <- funTypeDef
  eqs <- many (equation f)
  return $ Function f fty eqs

--- full programs

funProg :: GenParser Char u FunctionalProg
funProg = spaces *> return (,) <*> many dataDef <*> many funDef <* eof
  
parseProg :: String -> String -> CheckS FunctionalProg
parseProg fileName content = case parse funProg fileName content of
  Left e -> Left $ show e
  Right fp -> Right fp
  
parseTerm :: String -> CheckS Term
parseTerm s = case parse (spaces *> term S.empty <* eof) "" s of
  Left e -> Left $ show e
  Right t -> Right t
  